@@ -12,12 +12,7 @@ import shtab  # for completion magic
 
 def get_main_parser():
     parser = argparse.ArgumentParser(prog="pathcomplete")
-    parser.add_argument(
-        "-s",
-        "--print-completion-shell",
-        choices=["bash", "zsh"],
-        help="prints completion script",
-    )
+    shtab.add_argument_to(parser, ["-s", "--print-completion"])  # magic!
     # file & directory tab complete
     parser.add_argument("file", nargs="?").complete = shtab.FILE
     parser.add_argument("--dir", default=".").complete = shtab.DIRECTORY
@@ -27,10 +22,4 @@ def get_main_parser():
 if __name__ == "__main__":
     parser = get_main_parser()
     args = parser.parse_args()
-
-    # completion magic
-    shell = args.print_completion_shell
-    if shell:
-        print(shtab.complete(parser, shell=shell))
-    else:
-        print("received <file>=%r --dir=%r" % (args.file, args.dir))
+    print("received <file>=%r --dir=%r" % (args.file, args.dir))
